/*
 * Copyright IBM Corp. 2021
 *
 * libzpc is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */

/*
 * Build test for aes_gcm.h.
 */
#include "zpc/aes_gcm.h"
#include "zpc/aes_gcm.h"

int b_aes_gcm_not_empty;
